#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef long long int ticks_t;
volatile int foo = 0;

#if !(N_OPERATIONS >= 0)
#error "Must define N_OPERATIONS"
#elif N_OPERATIONS > 37
#warning "Warning: clang only unrolls the loop if N_OPERATIONS <= 37"
#endif

static const int N_MEASUREMENTS = 100;

static inline void start(unsigned int *cycles_low, unsigned int *cycles_high) {
  asm volatile ("LFENCE\n\t"
                "RDTSC\n\t"
                "mov %%edx, %0\n\t"
                "mov %%eax, %1\n\t": "=r" (*cycles_high), "=r" (*cycles_low)::
                "%rax", "%rbx", "%rcx", "%rdx");
}

static inline ticks_t stop(unsigned int *start_cycles_low, unsigned int *start_cycles_high) {
  unsigned end_cycles_low, end_cycles_high;
#if HAVE_RDTSCP == 1
  asm volatile("RDTSCP\n\t"
               "mov %%edx, %0\n\t"
               "mov %%eax, %1\n\t"
               "LFENCE\n\t": "=r" (end_cycles_high), "=r" (end_cycles_low):: "%rax",
               "%rbx", "%rcx", "%rdx");
#else
  asm volatile("LFENCE\n\t"
               "RDTSC\n\t"
               "mov %%edx, %0\n\t"
               "mov %%eax, %1\n\t"
               "LFENCE\n\t": "=r" (end_cycles_high), "=r" (end_cycles_low):: "%rax",
               "%rbx", "%rcx", "%rdx");
#endif
  ticks_t start_cycles = ((ticks_t) *start_cycles_high << 32) | *start_cycles_low;
  ticks_t end_cycles = ((ticks_t) end_cycles_high << 32) | end_cycles_low;
  return end_cycles - start_cycles;
}


static ticks_t measure(int n_ops) {
    unsigned int cycles_low;
    unsigned int cycles_high;

    start(&cycles_low, &cycles_high);
    for (int i = 0; i < n_ops; ++i) {
        foo = 3;
    }
    return stop(&cycles_low, &cycles_high);
}

static int intcmp(const void *aa, const void *bb) {
    const int *a = aa, *b = bb;
    return (*a < *b) ? -1 : (*a > *b);
}

static void print_five_point_summary(int n_operations, ticks_t results[], size_t n) {
    ticks_t sorted_results[n];
    memcpy(sorted_results, results, n * sizeof(ticks_t));
    qsort(sorted_results, n, sizeof(ticks_t), intcmp);

    size_t indices[5];
    for (int i = 0; i < 5; ++i) {
        indices[i] = i * (n-1) / 4;
    }
    printf("n_operations: %d, min (%zd): %lld, q1 (%zd): %lld, median (%zd): %lld, q3 (%zd): %lld, max (%zd): %lld\n",
            n_operations,
            indices[0], sorted_results[indices[0]],
            indices[1], sorted_results[indices[1]],
            indices[2], sorted_results[indices[2]],
            indices[3], sorted_results[indices[3]],
            indices[4], sorted_results[indices[4]]);
    fprintf(stderr, "%d, %lld, %lld, %lld, %lld, %lld\n",
            n_operations,
            sorted_results[indices[0]],
            sorted_results[indices[1]],
            sorted_results[indices[2]],
            sorted_results[indices[3]],
            sorted_results[indices[4]]);
}

int main() {
    ticks_t results[N_MEASUREMENTS];
    for (int i = 0; i < N_MEASUREMENTS; ++i) {
        results[i] = measure(N_OPERATIONS);
    }
    print_five_point_summary(N_OPERATIONS, results, N_MEASUREMENTS);

    return 0;
}
